
#ifndef MP_SLEEP_WRAPPER_H_
#define MP_SLEEP_WRAPPER_H_

#define ENABLE_CPP11_IMPL

#include <cstdint>

#ifdef ENABLE_CPP11_IMPL
#include <chrono>
#include <thread>
#else // ENABLE_CPP11_IMPL
#if defined(_WIN32) || defined(_MSC_VER) || defined(MINGW)
#include <Windows.h>
#else // defined(_WIN32) || defined(_MSC_VER) || defined(MINGW)
#include <unistd.h>
#ifdef ENABLE_HIGHRES_MODE
#include <time.h>
#endif // ENABLE_HIGHRES_MODE
#endif // defined(_WIN32) || defined(_MSC_VER) || defined(MINGW)
#endif // ENABLE_CPP11_IMPL

#include <string>

namespace multiplatform {
namespace time {

#ifdef ENABLE_CPP11_IMPL
inline bool ms_sleep(std::int64_t ms){ std::this_thread::sleep_for(std::chrono::milliseconds(ms)); return true; }
inline bool us_sleep(std::int64_t us){ std::this_thread::sleep_for(std::chrono::microseconds(us)); return true; }
inline bool ns_sleep(std::int64_t ns){ std::this_thread::sleep_for(std::chrono::nanoseconds(ns)); return true; }
#else // ENABLE_CPP11_IMPL
#if defined(_MSC_VER) || defined(MINGW)
inline bool ms_sleep(std::int64_t ms){ ::Sleep(ms); return true; }

#ifdef ENABLE_HIGHRES_MODE
inline bool ns_wait_timer(std::int64_t ns){
  HANDLE t;          // Timer handle
  LARGE_INTEGER dt;  // Due time
  if(!(t = CreateWaitableTimer(NULL, TRUE, NULL))){ return false; }
  dt.QuadPart = -ns;
  // After a lapse of time, start the timer which sends the signal only once.
  if(!SetWaitableTimer(t, &dt, 0, NULL, NULL, FALSE)){
    CloseHandle(t);
    return false;
  }
  // Wait for the arrival of the signal.
  WaitForSingleObject(t, INFINITE);
  CloseHandle(t);
  return true;
}
inline bool us_sleep(std::int64_t us){ return ns_wait_timer(us*1000); }
inline bool ns_sleep(std::int64_t ns){ return ns_wait_timer(ns); }
#else // ENABLE_HIGHRES_MODE
inline bool us_sleep(std::int64_t us){ ::Sleep(us/1000); return true; }
inline bool ns_sleep(std::int64_t ns){ ::Sleep(ns/1000000); return true; }
#endif // ENABLE_HIGHRES_MODE

#else // defined(_MSC_VER) || defined(MINGW)
inline bool ms_sleep(std::int64_t ms){ usleep(ms*1000); return true; }

#ifdef ENABLE_HIGHRES_MODE
inline bool ns_wait_timer(std::int64_t ns){
  struct timespec treq;
  clock_gettime(CLOCK_REALTIME, &treq);
  treq.tv_sec  += (time_t) ns/1000000000;
  treq.tv_nsec += ns%1000000000;
  if(treq.tv_nsec>=1000000000){ treq.tv_sec++; treq.tv_nsec-=1000000000; }
  int err = 0;
  do{
    err = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &treq, NULL);
  } while(err<0);
  return true;
}
inline bool us_sleep(std::int64_t us){ return ns_wait_timer(us*1000); }
inline bool ns_sleep(std::int64_t ns){ return ns_wait_timer(ns); }
#else // ENABLE_HIGHRES_MODE
inline bool us_sleep(std::int64_t us){ usleep(us); return true; }
inline bool ns_sleep(std::int64_t ns){ usleep(ns/1000); return true; }
#endif // ENABLE_HIGHRES_MODE

#endif // defined(_MSC_VER) || defined(MINGW)

#endif // ENABLE_CPP11_IMPL

} // namespace time
} // namespace multiplatform

#endif // MP_SLEEP_WRAPPER_H_

